#coding: utf-8

import licant.util
from licant.core import WrongAction

import sys
from optparse import OptionParser

import inspect

def routine_decorator(func=None, deps=None):
    if inspect.isfunction(func):
        licant.core.core.add(licant.core.Routine(func))
        return func

    else:
        def decorator(func):
            licant.core.core.add(licant.core.Routine(func, deps=deps))
            return func
            
        return decorator


def cli_argv_parse(argv):
    parser = OptionParser()
    parser.add_option("-d", "--debug", action="store_true",
                      default=False, help="print full system commands")
    parser.add_option("-j", "--threads", default=1,
                      help="amount of threads for executor")
    parser.add_option("-t", "--trace", action="store_true",
                      default=False, help="print trace information")

    parser.add_option("--printruntime", action="store_true", default=False)

    opts, args = parser.parse_args(argv)
    return opts, args


def cliexecute(default, colorwrap=False, argv=sys.argv[1:], core=licant.core.core):
    if colorwrap:
        print(licant.util.green("[start]"))

    opts, args = cli_argv_parse(argv)

    core.runtime["debug"] = opts.debug or opts.trace
    core.runtime["trace"] = opts.trace
    core.runtime["threads"] = int(opts.threads)

    if opts.printruntime:
        print("PRINT RUNTIME:", core.runtime)

    if len(args) == 0:
        if default is None:
            licant.util.error("default target isn't set")

        try:
            target = core.get(default)
            ret = target.invoke(target.default_action, critical=True)
        except licant.core.WrongAction as e:
            licant.util.error("Enough default action " + licant.util.yellow(
                target.default_action) + " in default target " + licant.util.yellow(_default))

    if len(args) == 1:
        fnd = args[0]
        if fnd in core.targets:
            try:
                target = core.get(fnd)
                ret = target.invoke(target.default_action, critical=True)
            except licant.core.WrongAction as e:
                licant.util.error("target.default_action")
        else:
            try:
                target = core.get(default)
                ret = target.invoke(fnd, critical=True)
            except licant.core.WrongAction as e:
                licant.util.error("Can't find routine " + licant.util.yellow(fnd) +
                                  ". Enough target or default target action with same name.")

    if len(args) == 2:
        try:
            target = licant.core.core.get(args[0])
            ret = target.invoke(args[1], critical=True)
        except licant.core.WrongAction as e:
            licant.util.error("Can't find action " + licant.util.yellow(args[1]) +
                              " in target " + licant.util.yellow(args[0]))

    if colorwrap:
        print(licant.util.yellow("[finish]"))



#coding: utf-8

from __future__ import print_function

import licant
from licant.core import Target, UpdatableTarget, UpdateStatus, core, subtree
from licant.cache import fcache
from licant.util import red, green, yellow, purple, quite
import threading
from licant.util import deprecated
import os
import sys

_rlock = threading.RLock()


def do_execute(target, rule, msgfield, prefix=None):
    def sprint(*args, **kwargs):
        with _rlock:
            print(*args, **kwargs)

    rule = rule.format(**target.__dict__)

    message = getattr(target, msgfield, None)

    if not core.runtime["debug"] and message is not None:
        if not isinstance(message, quite):
            if prefix is not None:
                sprint(prefix, message.format(**target.__dict__))
            else:
                sprint(message.format(**target.__dict__))

    else:
        sprint(rule)

    ret = os.system(rule)

    if target.isfile == True:
        target.update_info(target)

    return True if ret == 0 else False


class Executor:
    def __init__(self, rule, msgfield='message'):
        self.rule = rule
        self.msgfield = msgfield

    def __call__(self, target, **kwargs):
        return do_execute(target, self.rule, self.msgfield, **kwargs)


class MakeFileTarget(UpdatableTarget):
    def __init__(self, tgt, deps, **kwargs):
        UpdatableTarget.__init__(self, tgt, deps, **kwargs)

    def clean(self):
        stree = subtree(self.tgt)
        return stree.invoke_foreach(ops="clr", cond=if_file_and_exist)

    def makefile(self):
        return self.recurse_update()
        #stree = subtree(self.tgt)
        #stree.invoke_foreach(ops="dirkeep")
        #stree.reverse_recurse_invoke(
        #    ops="update_if_need", threads=core.runtime["threads"])


class FileTarget(MakeFileTarget):
    def __init__(self, tgt, deps, force=False, **kwargs):
        MakeFileTarget.__init__(self, tgt, deps, **kwargs)
        self.isfile = True
        self.force = force
        self.clrmsg = "DELETE {tgt}"
        self.default_action = "makefile"

    def update_info(self, _self):
        fcache.update_info(self.tgt)
        return True

    def mtime(self):
        curinfo = fcache.get_info(self.tgt)
        if curinfo.exist == False:
            return 0
        else:
            return curinfo.mtime

    def dirkeep(self):
        """Create directory tree for this file if needed."""
        dr = os.path.normpath(os.path.dirname(self.tgt))
        if (not os.path.exists(dr)):
            print("MKDIR %s" % dr)
            os.system("mkdir -p {0}".format(dr))
        return True

    def is_exist(self):
        curinfo = fcache.get_info(self.tgt)
        return curinfo.exist

    def warn_if_not_exist(self):
        """Print warn if file isn't exist"""
        info = fcache.get_info(self.tgt)
        if info.exist == False:
            print("Warn: file {} isn`t exist".format(purple(self.tgt)))

    def clr(self):
        """Delete this file."""
        do_execute(self, "rm -f {tgt}", "clrmsg")

    def self_need(self):
        if self.force or self.is_exist() == False:
            return True

        maxtime = 0
        for dep in self.get_deplist():
            if dep.mtime() > maxtime:
                maxtime = dep.mtime()

        if maxtime > self.mtime():
            return True

        return False

    def update(self):
        return self.invoke("build")


class FileSet(MakeFileTarget):
    """Virtual file target`s set.

    For link a set of file objects to the licant tree
    without depend`s overhead."""

    def __init__(self, tgt, targets):
        MakeFileTarget.__init__(
            self, tgt=tgt, deps=targets, default_action="makefile")
        self.targets = targets
        self.__mtime = None

    def self_need(self):
        return False

    def update(self):
        pass

    def mtime(self):
        if (self.__mtime is None):
            maxtime = 0
            for dep in self.get_deplist():
                if dep.mtime() > maxtime:
                    maxtime = dep.mtime()
            self.__mtime = maxtime

        return self.__mtime


def source(tgt, deps=[]):
    """Index source file by licant core."""
    target = FileTarget(
        build=lambda self: self.warn_if_not_exist(),
        deps=deps,
        tgt=tgt,
    )
    target.clr = None
    target.dirkeep = None
    target.update_status = UpdateStatus.Keeped
    core.add(target)


def copy(tgt, src, adddeps=[], message="COPY {src} {tgt}"):
    """Make the file copy target."""
    core.add(FileTarget(
        tgt=tgt,
        build=Executor("cp {src} {tgt}"),
        src=src,
        deps=[src] + adddeps,
        message=message
    ))


def fileset(tgt, targets):
    """Make a fileset."""
    core.add(FileSet(
        tgt=tgt,
        targets=targets
    ))


def if_file_and_exist(target):
    if not isinstance(target, FileTarget):
        return False

    curinfo = fcache.get_info(target.tgt)
    return curinfo.exist


from licant.util import red
from licant.scripter import scriptq

special = ["__script__", "__dir__"]


class Module:
    def __init__(self, name, script, dir, stack, **kwargs):
        self.name = name
        self.script = script
        self.stack = stack
        self.opts = kwargs
        self.opts["__script__"] = self.script
        self.opts["__dir__"] = dir


class VariantModule:
    def __init__(self):
        self.impls = {}

    def addimpl(self, impl, mod):
        self.impls[impl] = mod


class ModuleLibrary:
    def __init__(self):
        self.modules = {}

    def register(self, mod):
        if mod.name in self.modules:
            print("Attempt to register the module {} again".format(red(mod.name)))
            exit(-1)
        else:
            self.modules[mod.name] = mod

    def register_impl(self, mod, impl):
        if mod.name in self.modules:
            if not isinstance(self.modules[mod.name], VariantModule):
                print("Attempt to register the module {} again".format(red(mod.name)))
                exit(-1)
            else:
                varmod = self.modules[mod.name]
        else:
            varmod = VariantModule()
            self.modules[mod.name] = varmod

        varmod.addimpl(impl, mod)

    def get(self, name, impl=None):
        if not name in self.modules:
            print("The missing module {} was requested".format(red(name)))
            exit(-1)

        m = self.modules[name]
        if impl is None:
            if isinstance(m, VariantModule):
                print("Need implementation: {}".format(red(name)))
                exit(-1)
            else:
                return m
        else:
            if isinstance(m, Module):
                print("This modile have only one implementation: {}".format(red(name)))
                exit(-1)
            else:
                if impl in m.impls:
                    return m.impls[impl]
                else:
                    print("Unregistred implementation: {}".format(red(impl)))
                    exit(-1)


mlibrary = ModuleLibrary()


def module(name, impl=None, **kwargs):
    if impl is not None:
        implementation(name, impl, **kwargs)
        return
    mlibrary.register(Module(name, script=scriptq.last(
    ), dir=scriptq.curdir(), stack=list(scriptq.stack), **kwargs))


def implementation(name, impl, **kwargs):
    mlibrary.register_impl(Module(name, script=scriptq.last(
    ), dir=scriptq.curdir(), stack=list(scriptq.stack), **kwargs), impl=impl)


class submodule:
    def __init__(self, name, impl=None, addopts=None):
        if (isinstance(name, tuple)):
            self.name = name[0]
            self.impl = name[1]
            self.addopts = addopts
        elif (isinstance(name, submodule)):
            self.name = name.name
            self.impl = name.impl
            self.addopts = addopts
        else:
            self.name = name
            self.impl = impl
            self.addopts = addopts

    def __repr__(self):
        return "subm(" + self.name + ")"



from licant.scripter import scriptq
from licant.util import yellow

import os
import sys
import json

gpath = "/var/lib/licant"
lpath = os.path.expanduser("~/.licant")

libs = None


def merge_two_dicts(x, y):
    z = x.copy()   # start with x's keys and values
    z.update(y)    # modifies z with y's keys and values & returns None
    return z


def include(lib):
    global libs
    if libs is None:
        glibs = {}
        llibs = {}

        if os.path.exists(gpath):
            glibs = json.load(open(gpath))

        if os.path.exists(lpath):
            llibs = json.load(open(lpath))

        libs = merge_two_dicts(glibs, llibs)

    if not lib in libs:
        print("Unregistred library {}. Use licant-config utility or manually edit {} or {} file.".format(
            yellow(lib), yellow(lpath), yellow(gpath)))
        exit(-1)

    scriptq.execute(libs[lib])



from licant.core import Core, Target, UpdatableTarget, UpdateStatus
from licant.core import core as default_core

from licant.make import copy, fileset

from licant.cli import cliexecute as ex
from licant.cli import routine_decorator as routine

from licant.modules import module, submodule
from licant.util import error

import licant.scripter

__version__ = '1.0.1'


def execute(path):
    licant.scripter.scriptq.execute(path)


def execute_recursive(*argv, **kwargs):
    licant.scripter.scriptq.execute_recursive(*argv, **kwargs)


def about():
    return "I'm Licant"



#!/usr/bin/env python3

from setuptools import setup
import glob
import os

import licant

setup(
	name = 'licant',
	packages = ['licant'],
	version = licant.__version__,
	license='MIT',
	description = 'licant make system',
	author = 'Sorokin Nikolay',
	author_email = 'mirmikns@yandex.ru',
	url = 'https://github.com/mirmik/licant',
	keywords = ['testing', 'make'],
	classifiers = [],

	scripts = ["configurator/licant-config"],
#	package_data={'licant': [
#   	'templates/cxx/make.py',
#	  	'templates/cxx/main.cpp',
#	 	'templates/cxxgxx/make.py',
#		'templates/cxxgxx/main.cpp',
#    ]}
)
